import logging

import dspy
from dspy import Predict
from pydantic import BaseModel

from autoflow.knowledge_graph.types import GeneratedKnowledgeGraph

# Initialize logger
logger = logging.getLogger(__name__)


class EvaluateKnowledgeGraph(dspy.Signature):
    """
    Evaluate the differences between two knowledge graphs and provide scores for each entity and relationship,
    as well as a final score for the entire knowledge graph.

    Steps:
    1. Iterate over each entity in the expected knowledge graph
    2. For each expected entity, find the most similar entity in the actual knowledge graph
    3. Calculate the score (range from 0 to 1) for the entity based on the similarity
    4. Iterate over each relationship in the expected knowledge graph
    5. For each expected relationship, find the most similar relationship in the actual knowledge graph
    6. Calculate the score (range from 0 to 1) for the relationship based on the similarity
    7. Average all the scores of entities and relationships to get the final score

    Please only respond in JSON format.
    """

    actual_graph: GeneratedKnowledgeGraph = dspy.InputField(
        desc="The actual knowledge graph extracted from the text"
    )
    expected_graph: GeneratedKnowledgeGraph = dspy.InputField(
        desc="The expected knowledge graph"
    )
    score: float = dspy.OutputField(
        desc="The final score of the actual knowledge graph"
    )


class KGEvaluationResult(BaseModel):
    expected: GeneratedKnowledgeGraph
    actual: GeneratedKnowledgeGraph
    score: float


class KnowledgeGraphEvaluator(dspy.Module):
    def __init__(self, dspy_lm: dspy.LM):
        super().__init__()
        self.dspy_lm = dspy_lm
        self.program = Predict(EvaluateKnowledgeGraph)

    def forward(
        self,
        actual: GeneratedKnowledgeGraph,
        expected: GeneratedKnowledgeGraph,
    ) -> KGEvaluationResult:
        # Evaluate the knowledge graph using the provided language model
        with dspy.settings.context(lm=self.dspy_lm):
            prediction = self.program(actual_graph=actual, expected_graph=expected)
            return KGEvaluationResult(
                actual=actual,
                expected=expected,
                score=prediction.score,
            )



from abc import abstractmethod, ABC

from autoflow.storage.graph_store.base import GraphStore
from autoflow.storage.types import QueryBundle
from autoflow.knowledge_graph.types import RetrievedKnowledgeGraph


class KGRetriever(ABC):
    def __init__(self, knowledge_graph_store: GraphStore):
        self._kg_store = knowledge_graph_store

    @abstractmethod
    def retrieve(
        self,
        query: QueryBundle,
        depth: int = 2,
        meta_filters: dict = None,
    ) -> RetrievedKnowledgeGraph:
        raise NotImplementedError



import sqlalchemy
from pytidb.utils import build_tidb_dsn
from sqlalchemy import Engine

from autoflow.configs.db import DatabaseConfig


def get_db_engine_from_config(db_config: DatabaseConfig) -> Engine:
    if db_config.database_url is not None:
        database_url = db_config.database_url
    else:
        database_url = str(
            build_tidb_dsn(
                host=db_config.host,
                port=db_config.port,
                username=db_config.username,
                password=db_config.password,
                database=db_config.database,
                enable_ssl=db_config.enable_ssl,
            )
        )

    # Notice:
    # In order to save resource consumption, the tidb serverless cluster will "pause" automatically if there
    # are no active connections for more than 5 minutes, it will close all connections on the server side,
    # so we also need to recycle the connections from the connection pool on the client side.
    db_engine = sqlalchemy.create_engine(
        database_url,
        pool_size=20,
        max_overflow=40,
        pool_recycle=300,
        pool_pre_ping=True,
    )

    return db_engine



from typing import Optional

from pydantic import Field, BaseModel


class BaseLLMConfig(BaseModel):
    model: str = Field(
        description="The model to use for the LLM",
        default="gpt-4o",
    )
    max_tokens: Optional[int] = None
    temperature: float = 0.1



from uuid import UUID
from datetime import datetime
from typing import Optional, List, Any

from pydantic import BaseModel, Field, computed_field

from autoflow.data_types import DataType
from autoflow.utils import uuid6
from autoflow.utils.hash import sha256


# Chunk


class Chunk(BaseModel):
    id: Optional[UUID] = Field(default_factory=uuid6.uuid7)
    text: str = Field(description="The text of the chunk.")
    text_vec: Optional[Any] = Field(
        default=None, description="The vector of text vectors."
    )
    meta: Optional[dict] = Field(
        default_factory=dict, description="The metadata of the chunk."
    )
    document_id: Optional[UUID] = Field(
        default=None, description="The id of the document that the chunk belongs to."
    )
    created_at: datetime = Field(default=None, description="The created time")
    updated_at: datetime = Field(default=None, description="The updated time")

    @computed_field
    @property
    def hash(self) -> Optional[str]:
        return sha256(self.text)


class RetrievedChunk(Chunk):
    score: Optional[float] = Field(description="The score of the chunk.", default=None)
    similarity_score: Optional[float] = Field(
        default=None, description="The similarity score of the chunk."
    )


# Document


class Document(BaseModel):
    id: Optional[UUID] = Field(default_factory=uuid6.uuid7)
    name: Optional[str] = Field(None, description="The name of the document.")
    content: str = Field(description="The content of the document.")
    data_type: Optional[DataType] = Field(
        default=None, description="The data type of the document."
    )
    meta: Optional[dict] = Field(
        default_factory=dict, description="The metadata of the document."
    )
    created_at: Optional[datetime] = Field(default=None, description="The created time")
    updated_at: Optional[datetime] = Field(default=None, description="The updated time")
    chunks: Optional[List[Chunk]] = Field(
        default_factory=list, description="The chunks of the document."
    )

    @computed_field
    @property
    def hash(self) -> Optional[str]:
        return sha256(self.content)


class DocumentDescriptor(BaseModel):
    id: UUID
    name: str
    # source_uri: str


class DocumentSearchResult(BaseModel):
    chunks: List[RetrievedChunk] = Field(
        default_factory=list, description="The chunks of the search result."
    )
    documents: List[DocumentDescriptor | Document] = Field(
        default_factory=list,
        description="The aggregated documents of the search result.",
    )



from autoflow.configs.models.providers.base import (
    ModelProviders,
    ModelProviderInfo,
    ProviderConfig,
)

model_providers = [
    ModelProviderInfo(
        name=ModelProviders.OPENAI,
        display_name="OpenAI",
        description="The OpenAI API provides a simple interface for developers to create an intelligence layer in their applications, powered by OpenAI's state of the art models.",
        website="https://platform.openai.com",
        supported_model_types=["llm", "text_embedding"],
    )
]

model_provider_mappings = {provider.name: provider for provider in model_providers}

__all__ = [
    "ModelProviders",
    "ModelProviderInfo",
    "ProviderConfig",
    "model_providers",
    "model_provider_mappings",
]



import logging
from typing import Optional

import dspy

from autoflow.knowledge_graph.extractors.simple import SimpleKGExtractor
from autoflow.knowledge_graph.retrievers.weighted import WeightedGraphRetriever
from autoflow.knowledge_graph.types import (
    RetrievedKnowledgeGraph,
)
from autoflow.models.embedding_models import EmbeddingModel
from autoflow.storage.doc_store.types import Chunk
from autoflow.storage.graph_store.base import GraphStore
from autoflow.storage.graph_store.types import KnowledgeGraph
from autoflow.types import BaseComponent


logger = logging.getLogger(__name__)


class KnowledgeGraphIndex(BaseComponent):
    def __init__(
        self,
        kg_store: GraphStore,
        dspy_lm: dspy.LM,
        embedding_model: EmbeddingModel,
    ):
        super().__init__()
        self._kg_store = kg_store
        self._dspy_lm = dspy_lm
        self._embedding_model = embedding_model
        self._kg_extractor = SimpleKGExtractor(self._dspy_lm)

    def add_text(self, text: str) -> Optional[KnowledgeGraph]:
        knowledge_graph = self._kg_extractor.extract(text)
        return self._kg_store.add(knowledge_graph.to_create())

    def add_chunk(self, chunk: Chunk) -> Optional[KnowledgeGraph]:
        # Check if the chunk has been added.
        exists_relationships = self._kg_store.list_relationships(chunk_id=chunk.id)
        if len(exists_relationships) > 0:
            logger.warning(
                "The subgraph of chunk %s has already been added, skip.", chunk.id
            )
            return None

        logger.info("Extracting knowledge graph from chunk %s", chunk.id)
        knowledge_graph = self._kg_extractor.extract(chunk)
        logger.info("Knowledge graph extracted from chunk %s", chunk.id)

        return self._kg_store.add(knowledge_graph.to_create())

    def retrieve(
        self,
        query: str,
        depth: int = 2,
        metadata_filters: Optional[dict] = None,
        **kwargs,
    ) -> RetrievedKnowledgeGraph:
        retriever = WeightedGraphRetriever(
            self._kg_store,
            self._embedding_model,
            **kwargs,
        )
        return retriever.retrieve(
            query=query,
            depth=depth,
            metadata_filters=metadata_filters,
        )



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel
from autoflow.configs.models.llms import LLMConfig
from autoflow.configs.models.embeddings import EmbeddingModelConfig
from autoflow.configs.models.rerankers import RerankerConfig

# Index Methods


class IndexMethod(str, Enum):
    VECTOR_SEARCH = "VECTOR_SEARCH"
    FULLTEXT_SEARCH = "FULLTEXT_SEARCH"
    KNOWLEDGE_GRAPH = "KNOWLEDGE_GRAPH"


DEFAULT_INDEX_METHODS = [IndexMethod.VECTOR_SEARCH]

# Knowledge Base Config


class Version(int, Enum):
    V1 = 1


class KnowledgeBaseConfig(BaseModel):
    version: int = Version.V1
    name: str
    description: Optional[str] = None
    index_methods: List[IndexMethod] = DEFAULT_INDEX_METHODS
    llm: LLMConfig = None
    embedding_model: EmbeddingModelConfig = None
    reranker: RerankerConfig = None



import logging
from typing import List

import dspy
from dspy import Predict
from pydantic import BaseModel, Field

from autoflow.knowledge_graph.types import (
    GeneratedEntity,
    GeneratedKnowledgeGraph,
    GeneratedRelationship,
)

logger = logging.getLogger(__name__)


class PredictEntity(BaseModel):
    """Entity extracted from the text to form the knowledge graph"""

    name: str = Field(
        description="Name of the entity, it should be a clear and concise term"
    )
    description: str = Field(
        description=(
            "Description of the entity, it should be a complete and comprehensive sentence, not few words. "
            "Sample description of entity 'TiDB in-place upgrade': "
            "'Upgrade TiDB component binary files to achieve upgrade, generally use rolling upgrade method'"
        )
    )


class PredictRelationship(BaseModel):
    """Relationship extracted from the text to form the knowledge graph"""

    source_entity: str = Field(
        description="Source entity name of the relationship, it should an existing entity in the Entity list"
    )
    target_entity: str = Field(
        description="Target entity name of the relationship, it should an existing entity in the Entity list"
    )
    relationship_desc: str = Field(
        description=(
            "Description of the relationship, it should be a complete and comprehensive sentence, not few words. "
            "For example: 'TiDB will release a new LTS version every 6 months.'"
        )
    )


class PredictKnowledgeGraph(BaseModel):
    """Graph representation of the knowledge for text."""

    entities: List[PredictEntity] = Field(
        description="List of entities in the knowledge graph"
    )
    relationships: List[PredictRelationship] = Field(
        description="List of relationships in the knowledge graph"
    )

    def to_pandas(self):
        from pandas import DataFrame

        return {
            "entities": DataFrame(
                [
                    {
                        "name": entity.name,
                        "description": entity.description,
                    }
                    for entity in self.entities
                ]
            ),
            "relationships": DataFrame(
                [
                    {
                        "source_entity": relationship.source_entity,
                        "relationship_desc": relationship.relationship_desc,
                        "target_entity": relationship.target_entity,
                    }
                    for relationship in self.relationships
                ]
            ),
        }


class ExtractKnowledgeGraph(dspy.Signature):
    """Carefully analyze the provided text from database documentation and community blogs to thoroughly identify all entities related to database technologies, including both general concepts and specific details.

    Follow these Step-by-Step Analysis:

    1. Extract Meaningful Entities:
      - Identify all significant nouns, proper nouns, and technical terminologies that represent database-related concepts, objects, components, features, issues, key steps, execute order, user case, locations, versions, or any substantial entities.
      - Ensure that you capture entities across different levels of detail, from high-level overviews to specific technical specifications, to create a comprehensive representation of the subject matter.
      - Choose names for entities that are specific enough to indicate their meaning without additional context, avoiding overly generic terms.
      - Consolidate similar entities to avoid redundancy, ensuring each represents a distinct concept at appropriate granularity levels.

    2. Establish Relationships:
      - Carefully examine the text to identify all relationships between clearly-related entities, ensuring each relationship is correctly captured with accurate details about the interactions.
      - Analyze the context and interactions between the identified entities to determine how they are interconnected, focusing on actions, associations, dependencies, or similarities.
      - Clearly define the relationships, ensuring accurate directionality that reflects the logical or functional dependencies among entities. \
         This means identifying which entity is the source, which is the target, and what the nature of their relationship is (e.g., $source_entity depends on $target_entity for $relationship).

    Some key points to consider:
      - Please endeavor to extract all meaningful entities and relationships from the text, avoid subsequent additional gleanings.

    Objective: Produce a detailed and comprehensive knowledge graph that captures the full spectrum of entities mentioned in the text, along with their interrelations, reflecting both broad concepts and intricate details specific to the database domain.

    Please only response in JSON format.
    """

    text = dspy.InputField(
        desc="a paragraph of text to extract entities and relationships to form a knowledge graph"
    )
    knowledge: PredictKnowledgeGraph = dspy.OutputField(
        desc="Graph representation of the knowledge extracted from the text."
    )


class KnowledgeGraphExtractor(dspy.Module):
    def __init__(self, dspy_lm: dspy.LM):
        super().__init__()
        self.dspy_lm = dspy_lm
        self.program = Predict(ExtractKnowledgeGraph)

    def forward(self, text: str) -> GeneratedKnowledgeGraph:
        with dspy.settings.context(lm=self.dspy_lm):
            prediction = self.program(text=text)
            entities = [
                GeneratedEntity(
                    name=entity.name,
                    description=entity.description,
                    meta={},
                )
                for entity in prediction.knowledge.entities
            ]
            relationships = [
                GeneratedRelationship(
                    source_entity_name=relationship.source_entity,
                    target_entity_name=relationship.target_entity,
                    description=relationship.relationship_desc,
                    meta={},
                )
                for relationship in prediction.knowledge.relationships
            ]
            return GeneratedKnowledgeGraph(
                entities=entities,
                relationships=relationships,
            )



from uuid import UUID
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from autoflow.storage.doc_store.types import Chunk, Document, DocumentSearchResult


class DocumentStore(ABC):
    @abstractmethod
    def add(self, documents: List[Document]) -> List[Document]:
        raise NotImplementedError()

    @abstractmethod
    def update(self, document_id: UUID, update: Dict[str, Any]):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, document_id: UUID) -> None:
        raise NotImplementedError()

    @abstractmethod
    def list(self, filters: Dict[str, Any] = None) -> List[Document]:
        raise NotImplementedError()

    @abstractmethod
    def search(
        self,
        query: str,
        top_k: Optional[int] = None,
        similarity_candidate: Optional[int] = None,
    ) -> DocumentSearchResult:
        raise NotImplementedError()

    @abstractmethod
    def get(self, document_id: UUID) -> Document:
        raise NotImplementedError()

    @abstractmethod
    def add_doc_chunks(self, document_id: UUID, chunks: List[Chunk]) -> List[Chunk]:
        raise NotImplementedError()

    @abstractmethod
    def list_doc_chunks(self, document_id: UUID) -> List[Chunk]:
        raise NotImplementedError()

    @abstractmethod
    def get_chunk(self, chunk_id: UUID) -> Chunk:
        raise NotImplementedError()

    @abstractmethod
    def update_chunk(self, chunk_id: UUID, update: Dict[str, Any]) -> Chunk:
        raise NotImplementedError()

    @abstractmethod
    def delete_chunk(self, chunk_id: UUID) -> None:
        raise NotImplementedError()



from pydantic import BaseModel, Field


class BaseRerankerConfig(BaseModel):
    model: str = Field(
        description="The model to use for the reranker",
        default="jina-reranker-v2-base-multilingual",
    )
    top_n: int = Field(
        description="The number of results to return",
        default=5,
    )



from abc import abstractmethod
from typing import Generator

from autoflow.types import BaseComponent
from autoflow.storage.doc_store import Document


class Loader(BaseComponent):
    @abstractmethod
    def load(
        self, source: str | list[str], **kwargs
    ) -> Generator[Document, None, None]:
        raise NotImplementedError


class FileLoader(Loader):
    def load(self, files: str | list[str], **kwargs) -> Generator[Document, None, None]:
        if isinstance(files, str):
            files = [files]

        for file in files:
            yield self._load_file(file)

    @abstractmethod
    def _load_file(self, file: str) -> Document:
        raise NotImplementedError



from abc import abstractmethod

from autoflow.types import BaseComponent
from autoflow.storage.doc_store import Document


class Chunker(BaseComponent):
    @abstractmethod
    def chunk(self, document: Document) -> Document:
        raise NotImplementedError



from pydantic import BaseModel, Field


DEFAULT_CHUNK_SIZE = 1024  # tokens
DEFAULT_CHUNK_OVERLAP = 20  # tokens


class TextChunkerConfig(BaseModel):
    chunk_size: int = Field(default=DEFAULT_CHUNK_SIZE, description="Chunk size")
    chunk_overlap: int = Field(
        default=DEFAULT_CHUNK_OVERLAP, description="Chunk overlap"
    )



from autoflow.data_types import DataType
from autoflow.loaders.base import FileLoader
from autoflow.storage.doc_store import Document


class MarkdownLoader(FileLoader):
    def _load_file(self, file: str) -> Document:
        with open(file, "r", encoding="utf-8") as f:
            content = f.read()

        return Document(
            name=file,
            data_type=DataType.MARKDOWN,
            content=content,
        )



from autoflow.data_types import DataType
from autoflow.loaders import Loader


def get_loader_for_datatype(datatype: DataType) -> Loader:
    if datatype == DataType.MARKDOWN:
        from autoflow.loaders.markdown import MarkdownLoader

        return MarkdownLoader()
    elif datatype == DataType.PDF:
        from autoflow.loaders.pdf import PDFLoader

        return PDFLoader()
    elif datatype == DataType.HTML:
        from autoflow.loaders.webpage import WebpageLoader

        return WebpageLoader()
    else:
        raise ValueError(f"Unsupported loader for data type: {datatype}")



from typing import Optional

from pydantic import BaseModel, Field


class DatabaseConfig(BaseModel):
    database_url: Optional[str] = Field(
        description="Database connection string",
        default=None,
    )
    provider: Optional[str] = Field(
        description="Database provider",
        default="tidb",
    )
    host: Optional[str] = Field(
        description="Database host.",
        default="localhost",
    )
    port: Optional[int] = Field(description="Database port.", default=4000)
    username: Optional[str] = Field(
        description="The username to connect the database.",
        default="root",
    )
    password: Optional[str] = Field(
        description="The password to connect the database.",
        default="",
    )
    database: str = Field(
        description="Default name for the database",
        default="autoflow",
    )
    enable_ssl: Optional[bool] = Field(
        description="Enable SSL connection.",
        default=True,
    )



from typing import Dict

from pydantic import BaseModel, Field

from autoflow.configs.models.providers import ProviderConfig


class ManagerConfig(BaseModel):
    providers: Dict[str, ProviderConfig] = Field(default_factory=dict)



import logging
from datetime import datetime, UTC
from typing import Generator, Optional, List
from playwright.sync_api import sync_playwright
from bs4 import BeautifulSoup
from markdownify import MarkdownConverter

from autoflow.loaders.base import Loader
from autoflow.storage.doc_store import Document
from autoflow.data_types import DataType

logger = logging.getLogger(__name__)

# Common tags and classes to ignore when processing web content
IGNORE_TAGS = ["nav", "header", "footer", "script", "style", "noscript", "iframe"]
IGNORE_CLASSES = [
    "nav",
    "navigation",
    "footer",
    "header",
    "sidebar",
    "menu",
    "ad",
    "advertisement",
]


class WebpageLoader(Loader):
    def __init__(
        self,
        ignore_tags: Optional[List[str]] = None,
        ignore_classes: Optional[List[str]] = None,
    ):
        super().__init__()
        self._ignore_tags = ignore_tags or IGNORE_TAGS
        self._ignore_classes = ignore_classes or IGNORE_CLASSES

    def load(self, urls: str | list[str], **kwargs) -> Generator[Document, None, None]:
        if isinstance(urls, str):
            urls = [urls]

        visited = set()
        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            try:
                for url in urls:
                    try:
                        page = browser.new_page()
                        response = page.goto(url)
                        final_url = page.url

                        if final_url in visited:
                            continue

                        if response is None or response.status >= 400:
                            logger.error(
                                f"Failed to load page: {url}, response status: {response.status if response else 'None'}, skipping"
                            )
                            continue

                        # Parse the content
                        soup = BeautifulSoup(page.content(), "html.parser")

                        # Remove unwanted elements
                        for tag in self._ignore_tags:
                            for element in soup.find_all(tag):
                                element.extract()

                        for class_name in self._ignore_classes:
                            for element in soup.find_all(class_=class_name):
                                element.extract()

                        # Convert to markdown
                        content = MarkdownConverter().convert_soup(soup)
                        title = page.title() or final_url

                        visited.add(final_url)

                        # Create document
                        document = Document(
                            name=title,
                            content=content,
                            data_type=DataType.HTML,
                            meta={
                                "source_uri": final_url,
                                "original_uri": url,
                                "last_modified": datetime.now(UTC).isoformat(),
                            },
                        )

                        yield document

                    except Exception as e:
                        logger.error(f"Error processing URL {url}: {str(e)}")
                        continue
                    finally:
                        if "page" in locals():
                            page.close()
            finally:
                browser.close()



from enum import Enum
from typing import List, Optional, Dict, Any
from uuid import UUID
from datetime import datetime

from pydantic import BaseModel, model_validator, Field


# Entity


class EntityType(str, Enum):
    original = "original"
    synopsis = "synopsis"

    def __str__(self):
        return self.value


class Entity(BaseModel):
    id: Optional[UUID]
    entity_type: Optional[EntityType] = Field(
        description="Type of the entity", default=EntityType.original
    )
    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")
    embedding: Optional[Any] = Field(
        description="Embedding of the entity", default=None
    )
    meta: Optional[Dict[str, Any]] = Field(
        description="Metadata of the entity", default_factory=dict
    )
    created_at: Optional[datetime]
    updated_at: Optional[datetime]


class EntityCreate(BaseModel):
    entity_type: EntityType = EntityType.original
    name: Optional[str] = None
    description: Optional[str] = None
    meta: Optional[dict] = None


class SynopsisEntityCreate(EntityCreate):
    topic: str
    entities: List[int] = Field(description="The id list of the related entities")

    @model_validator(mode="after")
    def validate_entities(self):
        if len(self.entities) == 0:
            raise ValueError("Entities list should not be empty")
        return self


class EntityFilters(BaseModel):
    entity_type: Optional[EntityType] = None
    entity_id: Optional[UUID | List[UUID]] = None


class EntityUpdate(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    meta: Optional[dict] = None
    embedding: Optional[list[float]] = None


class EntityDegree(BaseModel):
    out_degree: int = 0
    in_degree: int = 0
    degrees: int = 0


# Relationship


class Relationship(BaseModel):
    id: Optional[UUID]
    source_entity_id: Optional[UUID] = Field(default=None)
    source_entity: Optional[Entity] = Field(default=None)
    target_entity_id: Optional[UUID] = Field(default=None)
    target_entity: Optional[Entity] = Field(default=None)
    description: str = Field(description="Description of the relationship")
    weight: Optional[float] = Field(default=0, description="Weight of the relationship")
    meta: Optional[Dict[str, Any]] = Field(
        description="Metadata of the relationship", default_factory=dict
    )
    embedding: Optional[Any] = Field(
        description="Embedding of the relationship", default=None
    )
    created_at: Optional[datetime] = Field(default=None)
    updated_at: Optional[datetime] = Field(default=None)


class RelationshipCreate(BaseModel):
    source_entity_name: str
    target_entity_name: str
    description: str
    meta: Dict[str, Any] = Field(default_factory=dict)
    weight: Optional[float] = Field(default=0)
    chunk_id: Optional[UUID] = Field(default=None)
    document_id: Optional[UUID] = Field(default=None)


class RelationshipUpdate(BaseModel):
    description: Optional[str] = None
    embedding: Optional[list[float]] = None


class RelationshipFilters(BaseModel):
    entity_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the entity connected to the relationship",
        default=None,
    )
    target_entity_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the target entity of the relationship",
        default=None,
    )
    source_entity_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the source entity of the relationship",
        default=None,
    )
    chunk_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the chunk which the relationship belongs to",
        default=None,
    )
    document_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the document which the relationship belongs to",
        default=None,
    )
    relationship_id: Optional[UUID | List[UUID]] = Field(
        description="Filter by the id of the relationship",
        default=None,
    )
    exclude_relationship_ids: Optional[List[UUID]] = Field(
        description="Exclude the relationships by the id",
        default=None,
    )
    metadata: Optional[Dict[str, Any]] = Field(
        description="Filter by the metadata of the relationship",
        default=None,
    )


# Knowledge Graph


class KnowledgeGraph(BaseModel):
    entities: List[Entity] = Field(default_factory=list)
    relationships: List[Relationship] = Field(default_factory=list)


# Knowledge Graph Create


class KnowledgeGraphCreate(BaseModel):
    entities: List[EntityCreate]
    relationships: List[RelationshipCreate]



from abc import ABC

import logging
from typing import (
    Collection,
    Dict,
    List,
    Optional,
    Tuple,
)
from uuid import UUID

from autoflow.storage.types import QueryBundle
from autoflow.types import BaseComponent
from autoflow.storage.graph_store.types import (
    Entity,
    EntityFilters,
    EntityType,
    EntityUpdate,
    EntityDegree,
    KnowledgeGraph,
    KnowledgeGraphCreate,
    Relationship,
    RelationshipFilters,
    RelationshipUpdate,
)


logger = logging.getLogger(__name__)


class GraphStore(BaseComponent, ABC):
    """Abstract base class for knowledge graph storage"""

    # Entity Basic Operations

    def list_entities(
        self, filters: Optional[EntityFilters] = EntityFilters()
    ) -> List[Entity]:
        """List all entities matching the filters"""
        raise NotImplementedError

    def search_entities(
        self,
        query: QueryBundle,
        top_k: int = 10,
        num_candidate: Optional[int] = None,
        distance_threshold: Optional[float] = None,
        filters: Optional[EntityFilters] = None,
    ) -> List[Tuple[Entity, float]]:
        raise NotImplementedError

    def get_entity(self, entity_id: UUID) -> Entity:
        """Get entity by ID"""
        raise NotImplementedError

    def must_get_entity(self, entity_id: UUID) -> Entity:
        entity = self.get_entity(entity_id)
        if entity is None:
            raise ValueError(f"Can not find the entity #{entity_id}")
        return entity

    def create_entity(
        self,
        name: str,
        entity_type: EntityType = EntityType.original,
        description: Optional[str] = None,
        meta: Optional[dict] = None,
        embedding: Optional[list[float]] = None,
    ) -> Entity:
        """Create a new entity"""
        raise NotImplementedError

    def update_entity(self, entity: Entity | UUID, update: EntityUpdate) -> Entity:
        """Update an existing entity"""
        raise NotImplementedError

    def delete_entity(self, entity_id: UUID) -> None:
        """Delete an entity"""
        raise NotImplementedError

    def delete_orphan_entities(self):
        """Remove entities that have no relationships"""
        raise NotImplementedError

    # Entity Degree Operations

    def calc_entity_out_degree(self, entity_id: UUID) -> int:
        """Calculate out-degree of an entity"""
        raise NotImplementedError

    def calc_entity_in_degree(self, entity_id: UUID) -> int:
        """Calculate in-degree of an entity"""
        raise NotImplementedError

    def calc_entity_degree(self, entity_id: UUID) -> int:
        """Calculate total degree of an entity"""
        raise NotImplementedError

    def calc_entities_degrees(
        self, entity_ids: Collection[UUID]
    ) -> Dict[UUID, EntityDegree]:
        """Calculate degrees for multiple entities in bulk"""
        raise NotImplementedError

    # Relationship Basic Operations

    def get_relationship(self, relationship_id: UUID) -> Relationship:
        """Get relationship by ID"""
        raise NotImplementedError

    def list_relationships(self, filters: RelationshipFilters) -> List[Relationship]:
        """List all relationships matching the filters"""
        raise NotImplementedError

    def create_relationship(
        self,
        source_entity: Entity,
        target_entity: Entity,
        description: Optional[str] = None,
        meta: Optional[dict] = {},
        **kwargs,
    ) -> Relationship:
        """Create a new relationship between entities"""
        raise NotImplementedError

    def update_relationship(
        self, relationship: Relationship | UUID, update: RelationshipUpdate
    ) -> Relationship:
        """Update an existing relationship"""
        raise NotImplementedError

    def delete_relationship(self, relationship_id: UUID):
        """Delete a relationship"""
        raise NotImplementedError

    def search_relationships(
        self,
        query: QueryBundle,
        top_k: int = 10,
        num_candidate: Optional[int] = None,
        distance_threshold: Optional[float] = None,
        distance_range: Optional[Tuple[float, float]] = None,
        filters: Optional[RelationshipFilters] = None,
    ) -> List[Tuple[Relationship, float]]:
        """

        Args:
            query:
            top_k:
            num_candidate:
            distance_threshold:
            distance_range:
            filters:
        """
        raise NotImplementedError

    def reset(self):
        """Reset the graph store"""
        raise NotImplementedError

    def drop(self):
        """Drop the graph store"""
        raise NotImplementedError

    # Knowledge Graph Operations

    def add(self, knowledge_graph: KnowledgeGraphCreate) -> Optional[KnowledgeGraph]:
        """Add a knowledge graph to the graph store"""
        raise NotImplementedError



import logging
from uuid import UUID
from typing import Any, Dict, List, Optional, Type

from pydantic import PrivateAttr

from pytidb import TiDBClient, Table
from pytidb.embeddings import EmbeddingFunction
from pytidb.schema import TableModel, Field, Column, Relationship as SQLRelationship
from pytidb.datatype import Vector, JSON
from pytidb.search import SearchType
from sqlalchemy.dialects.mysql import LONGTEXT

from autoflow.data_types import DataType
from autoflow.models.embedding_models import EmbeddingModel
from autoflow.orms.base import UUIDBaseModel
from autoflow.storage.doc_store.types import (
    Document,
    DocumentDescriptor,
    Chunk,
    RetrievedChunk,
    DocumentSearchResult,
)
from autoflow.types import SearchMode
from autoflow.storage.doc_store.base import DocumentStore


logger = logging.getLogger(__name__)


def dynamic_create_models(
    namespace: Optional[str] = None,
    embedding_model: Optional[EmbeddingModel] = None,
    vector_dims: Optional[int] = None,
) -> tuple[Type[TableModel], Type[TableModel]]:
    if embedding_model is None and vector_dims is None:
        raise ValueError("Either `embedding_model` or `vector_dims` must be specified")

    # Determine the table names.
    suffix = f"_{namespace}" if namespace else ""
    document_table_name = f"documents{suffix}"
    document_model_name = f"DBDocument{suffix}"
    chunk_table_name = f"chunks{suffix}"
    chunk_model_name = f"DBChunk{suffix}"

    # Initialize the document table model.
    class DBDocument(UUIDBaseModel):
        hash: str = Field(max_length=128)
        name: str = Field(max_length=256)
        content: str = Field(sa_column=Column(LONGTEXT))
        data_type: Optional[DataType] = Field(default=None)
        meta: dict = Field(default_factory=dict, sa_column=Column(JSON))

    document_model = type(
        document_model_name,
        (DBDocument,),
        {
            "__tablename__": document_table_name,
            "__table_args__": {"extend_existing": True},
        },
        table=True,
    )

    # Initialize the chunk table model.
    if embedding_model is not None:
        embed_fn = EmbeddingFunction(
            model_name=embedding_model.model_name,
            dimensions=embedding_model.dimensions,
            api_key=embedding_model.api_key,
            api_base=embedding_model.api_base,
            timeout=embedding_model.timeout,
        )
        vector_field = embed_fn.VectorField(source_field="text")
    else:
        vector_field = Field(default=None, sa_column=Column(Vector(vector_dims)))

    class DBChunk(UUIDBaseModel):
        text: str = Field(sa_column=Column(LONGTEXT))
        text_vec: Optional[Any] = vector_field
        document_id: UUID = Field(foreign_key=f"{document_table_name}.id")

    chunk_model = type(
        chunk_model_name,
        (DBChunk,),
        {
            "__tablename__": chunk_table_name,
            "__table_args__": {"extend_existing": True},
            "__annotations__": {
                "document": Optional[document_model],
            },
            "document": SQLRelationship(
                sa_relationship_kwargs={
                    "cascade": "all, delete",
                },
            ),
        },
        table=True,
    )

    return document_model, chunk_model


class TiDBDocumentStore(DocumentStore):
    _client: TiDBClient = PrivateAttr()
    _document_db_model: Type[Type[TableModel]] = PrivateAttr()
    _document_table: Table = PrivateAttr()
    _chunk_db_model: Type[Type[TableModel]] = PrivateAttr()
    _chunk_table: Table = PrivateAttr()

    def __init__(
        self,
        client: TiDBClient,
        namespace: Optional[str] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        vector_dims: Optional[int] = None,
    ) -> None:
        super().__init__()
        self._client = client
        self._db_engine = self._client.db_engine
        self._embedding_model = embedding_model
        self._init_store(namespace, vector_dims)

    @classmethod
    def class_name(cls) -> str:
        return "TiDBDocumentStore"

    def _init_store(
        self, namespace: Optional[str] = None, vector_dims: Optional[int] = None
    ):
        self._document_db_model, self._chunk_db_model = dynamic_create_models(
            namespace=namespace,
            vector_dims=vector_dims,
            embedding_model=self._embedding_model,
        )
        self._document_table = self._client.create_table(schema=self._document_db_model)
        self._chunk_table = self._client.create_table(schema=self._chunk_db_model)

    # Document Operations.

    def add(self, documents: List[Document]) -> List[Document]:
        """
        Add documents.
        """
        return_documents = []
        for doc in documents:
            db_document = self._document_db_model(**doc.model_dump(exclude={"chunks"}))
            db_document = self._document_table.insert(db_document)

            return_chunks = []
            if doc.chunks is not None and len(doc.chunks) > 0:
                db_chunks = self.add_doc_chunks(db_document.id, doc.chunks)
                return_chunks = [
                    Chunk(**db_chunk.model_dump(exclude={"document"}))
                    for db_chunk in db_chunks
                ]

            return_documents.append(
                Document(**db_document.model_dump(), chunks=return_chunks)
            )

        return return_documents

    def update(self, document_id: UUID, update: Dict[str, Any]) -> None:
        """
        Update documents.
        """
        self._document_table.update(update, {"id": document_id})

    def delete(self, document_id: UUID) -> None:
        """
        Delete document by id.

        Note: The related chunks will also be deleted by cascade.

        Args:
            document_id: The id of the document to delete.
        """
        return self._document_table.delete({"id": document_id})

    def get(self, document_id: UUID) -> Document:
        """
        Get document by id.
        """
        db_document = self._document_table.get(document_id)
        return Document(**db_document.model_dump())

    # TODO: Support pagination.
    def list(self, filters: Dict[str, Any] = None) -> List[Document]:
        """
        List all documents.
        """
        db_documents = self._document_table.query(filters)
        return [Document(**d.model_dump()) for d in db_documents]

    def search(
        self,
        query: str | List[float],
        mode: SearchMode = "vector",
        top_k: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
        num_candidate: Optional[int] = None,
        full_document: Optional[bool] = None,
    ) -> DocumentSearchResult:
        # TODO: Support Fulltext search.
        # TODO: Support Hybrid search.
        if mode != "vector":
            raise NotImplementedError(
                ".search() only supports vector search currently, fulltext and hybird search will be coming soon."
            )

        db_chunks = (
            self._chunk_table.search(query, query_type=SearchType.VECTOR_SEARCH)
            .distance_threshold(
                (1 - similarity_threshold) if similarity_threshold is not None else None
            )
            .num_candidate(num_candidate)
            .limit(top_k)
            .to_pydantic(with_score=True)
        )
        document_ids = [c.document_id for c in db_chunks]
        db_documents = self.list(
            {
                "id": {"$in": document_ids},
            }
        )
        return self._convert_to_retrieval_result(db_chunks, db_documents, full_document)

    def _convert_to_retrieval_result(
        self,
        db_chunks: List[TableModel],
        db_documents: List[TableModel],
        full_document: bool,
    ) -> DocumentSearchResult:
        return DocumentSearchResult(
            chunks=[
                RetrievedChunk(
                    **c.hit.model_dump(),
                    similarity_score=c.similarity_score,
                    score=c.score,
                )
                for c in db_chunks
            ],
            documents=[
                Document(**d.model_dump())
                if full_document
                else DocumentDescriptor(**d.model_dump())
                for d in db_documents
            ],
        )

    # Chunk Operations.

    def add_doc_chunks(self, document_id: UUID, chunks: List[Chunk]) -> List[Chunk]:
        """
        Add document chunks.
        """
        db_chunks = [
            self._chunk_db_model(
                **c.model_dump(exclude={"document_id"}), document_id=document_id
            )
            for c in chunks
        ]
        db_chunks = self._chunk_table.bulk_insert(db_chunks)
        return [Chunk(**c.model_dump(exclude={"document"})) for c in db_chunks]

    def list_doc_chunks(self, document_id: UUID) -> List[Chunk]:
        """
        List document chunks.
        """
        return self._chunk_table.query({"document_id": document_id})

    def get_chunk(self, chunk_id: UUID) -> Chunk:
        """
        Get chunk by id.
        """
        chunk = self._chunk_table.get(chunk_id)
        return Chunk(**chunk.model_dump(exclude={"document"}))

    def delete_chunk(self, chunk_id: UUID) -> None:
        """
        Delete document chunk.
        """
        return self._chunk_table.delete({"id": chunk_id})

    def update_chunk(self, chunk_id: UUID, update: Dict[str, Any]) -> Chunk:
        """
        Update chunk.
        """
        self._chunk_table.update(update, {"id": chunk_id})
        return self.get_chunk(chunk_id)

    # Document Store Operations.

    def recreate(self) -> None:
        self._client.drop_table(self._chunk_table.table_name)
        self._client.drop_table(self._document_table.table_name)
        self._document_table = self._client.create_table(schema=self._document_db_model)
        self._chunk_table = self._client.create_table(schema=self._chunk_db_model)

    def reset(self) -> None:
        with self._client.session():
            self._client.execute("SET FOREIGN_KEY_CHECKS = 0")
            self._chunk_table.truncate()
            self._document_table.truncate()
            self._client.execute("SET FOREIGN_KEY_CHECKS = 1")



import logging
from typing import List, Mapping, Any

import dspy
from dspy import Predict
from pydantic import BaseModel, Field

from autoflow.knowledge_graph.types import GeneratedEntity


logger = logging.getLogger(__name__)


class InputEntity(BaseModel):
    """List of entities extracted from the text to form the knowledge graph"""

    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")


class OutputEntity(BaseModel):
    """List of entities extracted from the text to form the knowledge graph"""

    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")
    covariates: Mapping[str, Any] = Field(
        description=(
            "The attributes (which is a comprehensive json TREE, the first field is always: 'topic') to claim the entity. "
        )
    )


class ExtractEntityCovariate(dspy.Signature):
    """Please carefully review the provided text and entities list which are already identified in the text.
    Focusing on identifying detailed covariates associated with each entities provided.

    Extract and link the covariates (which is a comprehensive json TREE, the first field is always: "topic") to their respective entities.
    Ensure all extracted covariates is clearly connected to the correct entity for accuracy and comprehensive understanding.
    Ensure that all extracted covariates are factual and verifiable within the text itself, without relying on external knowledge or assumptions.
    Collectively, the covariates should provide a thorough and precise summary of the entity's characteristics as described in the source material.

    Please only response in JSON format.
    """

    text = dspy.InputField(
        desc="a paragraph of text to extract covariates to claim the entities."
    )
    input: List[InputEntity] = dspy.InputField(
        desc="List of entities identified in the text."
    )
    output: List[OutputEntity] = dspy.OutputField(
        desc="List of entities with their covariates."
    )


class EntityCovariateExtractor(dspy.Module):
    def __init__(self, dspy_lm: dspy.LM):
        super().__init__()
        self.dspy_lm = dspy_lm
        self.program = Predict(ExtractEntityCovariate)

    def forward(
        self, text: str, entities: List[GeneratedEntity]
    ) -> List[GeneratedEntity]:
        with dspy.settings.context(lm=self.dspy_lm):
            input_entities = [
                InputEntity(
                    name=entity.name,
                    description=entity.description,
                )
                for entity in entities
            ]

            predict = self.program(
                text=text,
                input=input_entities,
            )

            output_entity_map = {entity.name: entity for entity in predict.output}
            for entity in entities:
                if entity.name in output_entity_map:
                    # Update the covariates in the metadata of the entity.
                    entity.meta = output_entity_map[entity.name].covariates

            return entities



import logging
import uuid
from typing import List, Optional, Any
from functools import partial
from os import cpu_count
from concurrent.futures import ThreadPoolExecutor

from pydantic import Field, PrivateAttr
from sqlalchemy import Engine
from llama_index.core.base.llms.types import ChatResponse

from autoflow.chunkers.base import Chunker
from autoflow.chunkers.helper import get_chunker_for_datatype
from autoflow.configs.knowledge_base import IndexMethod
from autoflow.data_types import DataType, guess_datatype
from autoflow.knowledge_base.prompts import QA_WITH_KNOWLEDGE_PROMPT_TEMPLATE
from autoflow.knowledge_graph.index import KnowledgeGraphIndex
from autoflow.loaders.base import Loader
from autoflow.loaders.helper import get_loader_for_datatype
from autoflow.models.llms import LLM
from autoflow.models.embedding_models import EmbeddingModel
from autoflow.models.llms.dspy import get_dspy_lm_by_llm
from autoflow.models.rerank_models import RerankModel
from autoflow.types import BaseComponent, SearchMode
from autoflow.storage.doc_store import DocumentSearchResult, Document

logger = logging.getLogger(__name__)


class KnowledgeBase(BaseComponent):
    _llm: LLM = PrivateAttr()
    namespace: Optional[str] = Field(default=None)
    name: Optional[str] = Field(default=None)
    description: Optional[str] = Field(default=None)
    index_methods: List[IndexMethod] = Field(default=[IndexMethod.VECTOR_SEARCH])

    def __init__(
        self,
        db_engine: Engine = None,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        index_methods: Optional[List[IndexMethod]] = None,
        llm: Optional[LLM] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        rerank_model: Optional[RerankModel] = None,
        max_workers: Optional[int] = None,
    ):
        super().__init__(
            namespace=namespace,
            name=name,
            description=description,
            index_methods=index_methods,
        )
        self._db_engine = db_engine
        self._llm = llm
        self._embedding_model = embedding_model
        self._reranker_model = rerank_model
        self._init_stores()
        self._init_indexes()
        self._max_workers = max_workers or cpu_count()

    def _init_stores(self):
        from autoflow.storage.doc_store.tidb_doc_store import TiDBDocumentStore
        from autoflow.storage.graph_store.tidb_graph_store import TiDBGraphStore
        from pytidb import TiDBClient

        self._tidb_client = TiDBClient(self._db_engine)
        self._doc_store = TiDBDocumentStore(
            client=self._tidb_client,
            embedding_model=self._embedding_model,
            namespace=self.namespace,
        )
        self._kg_store = TiDBGraphStore(
            client=self._tidb_client,
            embedding_model=self._embedding_model,
            namespace=self.namespace,
        )

    def _init_indexes(self):
        self._dspy_lm = get_dspy_lm_by_llm(self._llm)
        self._kg_index = KnowledgeGraphIndex(
            kg_store=self._kg_store,
            dspy_lm=self._dspy_lm,
            embedding_model=self._embedding_model,
        )

    def class_name(self):
        return "KnowledgeBase"

    def documents(self):
        return self._doc_store;

    def knowledge_graph(self):
        return self._kg_store

    def add(
        self,
        source: str | list[str],
        data_type: Optional[DataType] = None,
        loader: Optional[Loader] = None,
        chunker: Optional[Chunker] = None,
    ) -> List[Document]:
        if data_type is None:
            data_type = guess_datatype(source)
        if data_type is None:
            raise ValueError("Please provide a valid data type.")

        if loader is None:
            loader = get_loader_for_datatype(data_type)

        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            build_index_for_document = partial(
                self.build_index_for_document, chunker=chunker
            )

            results = executor.map(build_index_for_document, loader.load(source))

        return_documents = []
        for result in results:
            return_documents.append(result)
        return return_documents

    def build_index_for_document(
        self,
        document: Document,
        chunker: Optional[Chunker] = None,
    ) -> List[Document]:
        """
        Build index for a document.

        Args:
            document: The document to build index for.
            chunker: The chunker to use to chunk the document.

        Returns:
            A list of documents that are the result of indexing the original document.
        """
        # TODO: handle duplicate documents.
        if chunker is None:
            chunker = get_chunker_for_datatype(document.data_type)

        chunked_document = chunker.chunk(document)
        self.add_document(chunked_document)

        if IndexMethod.KNOWLEDGE_GRAPH in self.index_methods:

            def add_chunk_to_kg(chunk):
                logger.info("Adding chunk <id: %s> to knowledge graph.", chunk.id)
                self._kg_index.add_chunk(chunk)

            with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
                list(executor.map(add_chunk_to_kg, chunked_document.chunks))

        return chunked_document

    # Document management.

    def add_document(self, document: Document):
        self._doc_store.add([document])

    def add_documents(self, documents: List[Document]):
        return self._doc_store.add(documents)

    def list_documents(self) -> List[Document]:
        return self._doc_store.list()

    def get_document(self, doc_id: uuid.UUID) -> Document:
        return self._doc_store.get(doc_id)

    def delete_document(self, doc_id: uuid.UUID) -> None:
        return self._doc_store.delete(doc_id)

    # Search

    def search(self):
        # TODO: Support one interface search documents and knowledge graph at the same time.
        raise NotImplementedError()

    def search_documents(
        self,
        query: str,
        mode: SearchMode = "vector",
        similarity_threshold: Optional[float] = None,
        num_candidate: Optional[int] = None,
        top_k: Optional[int] = 5,
        **kwargs: Any,
    ) -> DocumentSearchResult:
        return self._doc_store.search(
            query=query,
            mode=mode,
            similarity_threshold=similarity_threshold,
            num_candidate=num_candidate,
            top_k=top_k,
            **kwargs,
        )

    def search_knowledge_graph(
        self,
        query: str,
        depth: int = 2,
        metadata_filters: Optional[dict] = None,
        **kwargs,
    ):
        return self._kg_index.retrieve(
            query=query,
            depth=depth,
            metadata_filters=metadata_filters,
            **kwargs,
        )

    # Generation.

    def ask(self, question: str) -> ChatResponse:
        result = self.search_documents(
            query=question,
            similarity_threshold=0.4,
            top_k=5,
        )
        chunks = result.chunks
        knowledge_graph = self.search_knowledge_graph(
            query=question,
        )
        messages = QA_WITH_KNOWLEDGE_PROMPT_TEMPLATE.format_messages(
            llm=self._llm,
            query_str=question,
            chunks=chunks,
            knowledge_graph=knowledge_graph,
        )
        return self._llm.chat(messages)

    # Knowledge Base Operation.

    def reset(self):
        self._doc_store.reset()
        self._kg_store.reset()



from enum import Enum
import os
from typing import IO, Optional, Union, BinaryIO, TextIO
from urllib.parse import urlparse


class DataType(str, Enum):
    MARKDOWN = "markdown"
    PDF = "pdf"
    DOCX = "docx"
    PPTX = "pptx"
    XLSX = "xlsx"
    CSV = "csv"
    SITEMAP = "sitemap"
    HTML = "html"


def guess_datatype(source: Union[str, IO, BinaryIO, TextIO]) -> Optional[DataType]:
    if isinstance(source, str):
        url = urlparse(source)
        if url.scheme == "" or url.scheme == "file":
            return guess_by_filename(url.path)
        elif url.scheme == "http" or url.scheme == "https":
            return DataType.HTML
        else:
            if os.path.exists(source):
                return guess_by_filename(source)
            raise ValueError(f"Unsupported URL scheme: {url.scheme}")
    elif isinstance(source, IO):
        return guess_by_filename(source.name)
    else:
        return None


def guess_by_filename(filename: str) -> Optional[DataType]:
    """Helper function to guess data type from filename."""
    lower = filename.lower()
    if lower.endswith(".md"):
        return DataType.MARKDOWN
    elif lower.endswith(".pdf"):
        return DataType.PDF
    elif lower.endswith(".docx"):
        return DataType.DOCX
    elif lower.endswith(".pptx"):
        return DataType.PPTX
    elif lower.endswith(".xlsx"):
        return DataType.XLSX
    elif lower.endswith(".csv"):
        return DataType.CSV
    elif lower.endswith(".xml") and "sitemap" in lower:
        return DataType.SITEMAP
    elif lower.endswith((".html", ".htm")):
        return DataType.HTML
    else:
        return None



from abc import abstractmethod

from autoflow.types import BaseComponent
from autoflow.knowledge_graph.types import GeneratedKnowledgeGraph


class KGExtractor(BaseComponent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def extract(self, text: str) -> GeneratedKnowledgeGraph:
        raise NotImplementedError()



from autoflow.chunkers.base import Chunker
from autoflow.data_types import DataType


def get_chunker_for_datatype(datatype: DataType) -> Chunker:
    if datatype in [DataType.MARKDOWN, DataType.HTML, DataType.PDF]:
        from autoflow.chunkers.text import TextChunker

        return TextChunker()
    else:
        raise ValueError(f"Unsupported data type: {datatype}")



from typing import Optional

from autoflow.chunkers.base import Chunker
from autoflow.configs.chunkers.text import TextChunkerConfig
from autoflow.storage.doc_store import Document, Chunk


class TextChunker(Chunker):
    """Chunker for text."""

    def __init__(self, config: Optional[TextChunkerConfig] = TextChunkerConfig()):
        super().__init__()
        from llama_index.core.node_parser import SentenceSplitter

        self._splitter = SentenceSplitter(
            chunk_size=config.chunk_size,
            chunk_overlap=config.chunk_overlap,
        )

    def chunk(self, document: Document) -> Document:
        texts = self._splitter.split_text(document.content)
        document.chunks = [Chunk(text=text) for text in texts]
        return document



from llama_index.readers import PDFReader

from autoflow.data_types import DataType
from autoflow.loaders.base import FileLoader
from autoflow.storage.doc_store import Document


class PDFLoader(FileLoader):
    def _load_file(self, file: str) -> Document:
        reader = PDFReader()
        documents = reader.load_data(file)
        content = documents[0].text

        return Document(
            name=file.name,
            data_type=DataType.PDF,
            content=content,
        )



import logging
from typing import Collection, Dict, List, Optional, Tuple, Type, Any
from uuid import UUID

from pydantic import PrivateAttr
from pytidb import Table, TiDBClient
from pytidb.datatype import JSON, Text
from pytidb.schema import (
    Column,
    Field,
    Relationship as SQLRelationship,
    TableModel,
    VectorField,
)
from pytidb.sql import func, select, or_
from pytidb.embeddings import EmbeddingFunction
from sqlalchemy import Index

from autoflow.models.embedding_models import EmbeddingModel
from autoflow.orms.base import UUIDBaseModel
from autoflow.storage.graph_store.base import GraphStore
from autoflow.storage.graph_store.types import (
    Entity,
    EntityDegree,
    EntityFilters,
    EntityType,
    EntityUpdate,
    KnowledgeGraph,
    KnowledgeGraphCreate,
    Relationship,
    RelationshipFilters,
    RelationshipUpdate,
)
from autoflow.storage.types import QueryBundle


logger = logging.getLogger(__name__)


def dynamic_create_models(
    namespace: Optional[str] = None,
    embedding_model: Optional[EmbeddingModel] = None,
    vector_dims: Optional[int] = None,
) -> tuple[Type[TableModel], Type[TableModel]]:
    if embedding_model is None and vector_dims is None:
        raise ValueError("Either `embedding_model` or `vector_dims` must be specified")

    # Determine the table names.
    suffix = f"_{namespace}" if namespace else ""
    entity_table_name = f"entities{suffix}"
    relationship_table_name = f"relationships{suffix}"
    entity_model_name = f"DBEntity{suffix}"
    relationship_model_name = f"DBRelationship{suffix}"

    # Embedding function.
    embed_fn = None
    if embedding_model is not None:
        embed_fn = EmbeddingFunction(
            model_name=embedding_model.model_name,
            dimensions=embedding_model.dimensions,
            api_key=embedding_model.api_key,
            api_base=embedding_model.api_base,
            timeout=embedding_model.timeout,
        )

    # Initialize the entity model.
    if embed_fn is not None:
        entity_vector_field = embed_fn.VectorField()
    else:
        entity_vector_field = VectorField(vector_dims)

    class DBEntity(UUIDBaseModel):
        __table_args__ = (
            Index("idx_entity_type", "entity_type"),
            Index("idx_entity_name", "name"),
        )
        entity_type: EntityType = EntityType.original
        name: str = Field(max_length=512)
        description: str = Field(sa_column=Column(Text))
        meta: Optional[Dict] = Field(default_factory=dict, sa_column=Column(JSON))
        embedding: Optional[Any] = entity_vector_field

        def __hash__(self):
            return hash(self.id)

        def __eq__(self, other):
            return self.id == other.id

    entity_model = type(
        entity_model_name,
        (DBEntity,),
        {
            "__tablename__": entity_table_name,
            "__table_args__": {
                "extend_existing": True,
            },
        },
        table=True,
    )

    # Initialize the relationship model.
    if embed_fn is not None:
        relationship_vector_field = embed_fn.VectorField()
    else:
        relationship_vector_field = VectorField(vector_dims)

    class DBRelationship(UUIDBaseModel):
        description: str = Field(sa_column=Column(Text))
        source_entity_id: UUID = Field(foreign_key=f"{entity_table_name}.id")
        target_entity_id: UUID = Field(foreign_key=f"{entity_table_name}.id")
        meta: Optional[Dict] = Field(default_factory=dict, sa_column=Column(JSON))
        embedding: Optional[Any] = relationship_vector_field
        weight: Optional[float] = Field(default=0)
        chunk_id: Optional[UUID] = Field(default=None)
        document_id: Optional[UUID] = Field(default=None)

        def __hash__(self):
            return hash(self.id)

        def __eq__(self, other):
            return self.id == other.id

    relationship_model = type(
        relationship_model_name,
        (DBRelationship,),
        {
            "__tablename__": relationship_table_name,
            "__table_args__": {"extend_existing": True},
            "__annotations__": {
                "source_entity": entity_model,
                "target_entity": entity_model,
            },
            "source_entity": SQLRelationship(
                sa_relationship_kwargs={
                    "primaryjoin": f"{relationship_model_name}.source_entity_id == {entity_model_name}.id",
                    "lazy": "joined",
                },
            ),
            "target_entity": SQLRelationship(
                sa_relationship_kwargs={
                    "primaryjoin": f"{relationship_model_name}.target_entity_id == {entity_model_name}.id",
                    "lazy": "joined",
                },
            ),
        },
        table=True,
    )

    return entity_model, relationship_model


class TiDBGraphStore(GraphStore):
    _db: TiDBClient = PrivateAttr()
    _entity_db_model: Type[TableModel] = PrivateAttr()
    _entity_table: Table = PrivateAttr()
    _relationship_db_model: Type[TableModel] = PrivateAttr()
    _relationship_table: Table = PrivateAttr()

    def __init__(
        self,
        client: TiDBClient,
        namespace: Optional[str] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        vector_dims: Optional[int] = None,
        entity_distance_threshold: Optional[float] = 0.1,
    ):
        super().__init__()
        self._db = client
        self._db_engine = client.db_engine
        self._embedding_model = embedding_model
        self._entity_distance_threshold = entity_distance_threshold
        self._init_store(namespace, vector_dims)

    def _init_store(
        self, namespace: Optional[str] = None, vector_dims: Optional[int] = None
    ):
        self._entity_db_model, self._relationship_db_model = dynamic_create_models(
            namespace=namespace,
            vector_dims=vector_dims,
            embedding_model=self._embedding_model,
        )
        self._entity_table = self._db.create_table(schema=self._entity_db_model)
        self._relationship_table = self._db.create_table(
            schema=self._relationship_db_model
        )

    # Entity Basic Operations

    def get_entity(self, entity_id: UUID) -> Entity:
        return self._entity_table.get(entity_id)

    def list_entities(
        self, filters: Optional[EntityFilters] = EntityFilters(), **kwargs
    ) -> List[Entity]:
        if isinstance(kwargs, dict):
            filters = filters.model_copy(update=kwargs)
        filter_dict = self._convert_entity_filters(filters)
        return self._entity_table.query(filter_dict)

    def search_entities(
        self,
        query: QueryBundle,
        top_k: int = 10,
        num_candidate: Optional[int] = None,
        distance_threshold: Optional[float] = None,
        filters: Optional[EntityFilters] = None,
    ) -> List[Tuple[Entity, float]]:
        filter_dict = self._convert_entity_filters(filters)
        results = (
            self._entity_table.search(query.query_embedding or query.query_str)
            .num_candidate(num_candidate or top_k * 10)
            .filter(filter_dict)
            .distance_threshold(distance_threshold)
            .limit(top_k)
            .to_pydantic()
        )
        return [(item.hit, item.score) for item in results]

    def _convert_entity_filters(self, filters: Optional[EntityFilters]) -> dict:
        filter_dict = {}
        if filters is None:
            return filter_dict
        if filters.entity_type:
            filter_dict["entity_type"] = filters.entity_type.value
        if filters.entity_id:
            op = "$in" if isinstance(filters.entity_id, list) else "$eq"
            filter_dict["id"] = {op: filters.entity_id}
        return filter_dict

    def create_entity(
        self,
        name: str,
        entity_type: EntityType = EntityType.original,
        description: Optional[str] = None,
        meta: Optional[dict] = None,
        embedding: Optional[list[float]] = None,
    ) -> Entity:
        if embedding is None:
            embedding = self._get_entity_embedding(name, description)
        entity = self._entity_db_model(
            name=name,
            entity_type=entity_type,
            description=description,
            meta=meta,
            embedding=embedding,
        )
        return self._entity_table.insert(entity)

    def _get_entity_embedding(self, name: str, description: str) -> list[float]:
        embedding_str = f"{name}: {description}"
        return self._embedding_model.get_text_embedding(embedding_str)

    def find_or_create_entity(
        self,
        name: str,
        entity_type: EntityType = EntityType.original,
        description: Optional[str] = None,
        meta: Optional[dict] = None,
        embedding: Optional[Any] = None,
    ) -> Entity:
        query_embedding = self._get_entity_embedding(name, description)
        query = QueryBundle(query_embedding=query_embedding)
        nearest_entity = self.search_entities(
            query, top_k=1, distance_threshold=self._entity_distance_threshold
        )
        if len(nearest_entity) != 0:
            return nearest_entity[0][0]
        else:
            return self.create_entity(
                name=name,
                entity_type=entity_type,
                description=description,
                meta=meta,
                embedding=embedding,
            )

    def update_entity(self, entity: Entity | UUID, update: EntityUpdate) -> Entity:
        if isinstance(entity, UUID):
            entity = self.get_entity(entity)

        update_dict = update.model_dump(exclude_none=True)
        if update.embedding is None:
            update_dict["embedding"] = self._get_entity_embedding(
                entity.name, entity.description
            )

        self._entity_table.update(values=update_dict, filters={"id": entity.id})
        # FIXME: pytidb should return the updated entity.
        entity = self._entity_table.get(entity.id)

        # Update connected relationships.
        connected_relationships = self.list_relationships(
            filters=RelationshipFilters(
                entity_id=entity.id,
            )
        )
        for relationship in connected_relationships:
            self.update_relationship(relationship, RelationshipUpdate())

        return entity

    def delete_entity(self, entity_id: UUID) -> None:
        with self._db.session():
            # Delete all relationships connected to the entity.
            self._relationship_table.delete({"source_entity_id": entity_id})
            self._relationship_table.delete({"target_entity_id": entity_id})

            # Delete the entity.
            self._entity_table.delete({"id": entity_id})

    def delete_orphan_entities(self):
        raise NotImplementedError()

    # Entity Degree Operations

    def calc_entity_out_degree(self, entity_id: UUID) -> int:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            self._relationship_db_model.source_entity_id == entity_id
        )
        return self._db.query(stmt).scalar()

    def calc_entity_in_degree(self, entity_id: UUID) -> int:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            self._relationship_db_model.target_entity_id == entity_id
        )
        return self._db.query(stmt).scalar()

    def calc_entity_degree(self, entity_id: UUID) -> int:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            or_(
                self._relationship_db_model.target_entity_id == entity_id,
                self._relationship_db_model.source_entity_id == entity_id,
            )
        )
        return self._db.query(stmt).scalar()

    def calc_entities_degrees(
        self, entity_ids: Collection[UUID]
    ) -> Dict[UUID, EntityDegree]:
        entity_table_name = self._entity_table.table_name
        relationship_table_name = self._relationship_table.table_name
        stmt = f"""
            SELECT
                e.id as id,
                COALESCE(SUM(CASE WHEN r.target_entity_id = e.id THEN 1 ELSE 0 END), 0) AS in_degree,
                COALESCE(SUM(CASE WHEN r.source_entity_id = e.id THEN 1 ELSE 0 END), 0) AS out_degree,
                COALESCE(COUNT(e.id), 0) AS degree
            FROM {entity_table_name} e
            LEFT JOIN {relationship_table_name} r ON e.id = r.source_entity_id OR e.id = r.target_entity_id
            WHERE e.id IN :entity_ids
            GROUP BY e.id
        """
        results = self._db.query(
            stmt, {"entity_ids": [entity_id.hex for entity_id in entity_ids]}
        ).to_list()
        return {
            UUID(item["id"]): EntityDegree(
                in_degree=item["in_degree"],
                out_degree=item["out_degree"],
                degrees=item["degree"],
            )
            for item in results
        }

    # Relationship Basic Operations

    def get_relationship(self, relationship_id: UUID) -> Relationship:
        return self._relationship_table.get(relationship_id)

    def list_relationships(
        self, filters: RelationshipFilters = RelationshipFilters(), **kwargs
    ) -> List[Relationship]:
        if isinstance(kwargs, dict):
            filters = filters.model_copy(update=kwargs)
        filter_dict = self._convert_relationship_filters(filters)
        return self._relationship_table.query(filter_dict)

    def search_relationships(
        self,
        query: QueryBundle,
        top_k: int = 10,
        num_candidate: Optional[int] = None,
        distance_threshold: Optional[float] = None,
        distance_range: Optional[Tuple[float, float]] = None,
        filters: Optional[RelationshipFilters] = None,
    ) -> List[Tuple[Relationship, float]]:
        filter_dict = self._convert_relationship_filters(filters)
        results = (
            self._relationship_table.search(query.query_embedding or query.query_str)
            .num_candidate(num_candidate or top_k * 10)
            .filter(filter_dict)
            .distance_threshold(distance_threshold)
            .distance_range(distance_range[0], distance_range[1])
            .limit(top_k)
            .to_pydantic()
        )

        # FIXME: pytidb should return the relationship field: target_entity, source_entity.
        entity_ids = [item.hit.target_entity_id for item in results]
        entity_ids.extend([item.hit.source_entity_id for item in results])
        entities = self.list_entities(filters=EntityFilters(entity_id=entity_ids))
        entity_map = {entity.id: entity for entity in entities}
        for item in results:
            item.hit.target_entity = entity_map[item.hit.target_entity_id]
            item.hit.source_entity = entity_map[item.hit.source_entity_id]

        return [(item.hit, item.score) for item in results]

    def _convert_relationship_filters(self, filters: RelationshipFilters) -> dict:
        filter_dict = {}

        if filters.entity_id:
            if isinstance(filters.entity_id, list):
                if len(filters.entity_id) != 0:
                    filter_dict["$or"] = [
                        {"target_entity_id": {"$in": filters.entity_id}},
                        {"source_entity_id": {"$in": filters.entity_id}},
                    ]
            else:
                filter_dict["$or"] = [
                    {"target_entity_id": {"$eq": filters.entity_id}},
                    {"source_entity_id": {"$eq": filters.entity_id}},
                ]

        if filters.source_entity_id:
            if isinstance(filters.source_entity_id, list):
                if len(filters.source_entity_id) != 0:
                    filter_dict["$or"] = [
                        {"source_entity_id": {"$in": filters.source_entity_id}}
                    ]
            else:
                filter_dict["$or"] = [
                    {"source_entity_id": {"$eq": filters.source_entity_id}}
                ]

        if filters.target_entity_id:
            if isinstance(filters.target_entity_id, list):
                if len(filters.target_entity_id) != 0:
                    filter_dict["$or"] = [
                        {"target_entity_id": {"$in": filters.target_entity_id}}
                    ]
            else:
                filter_dict["$or"] = [
                    {"target_entity_id": {"$eq": filters.target_entity_id}}
                ]

        if filters.relationship_id:
            if isinstance(filters.relationship_id, list):
                if len(filters.relationship_id) != 0:
                    filter_dict["id"] = {"$in": filters.relationship_id}
            else:
                filter_dict["id"] = {"$eq": filters.relationship_id}

        if (
            filters.exclude_relationship_ids
            and len(filters.exclude_relationship_ids) != 0
        ):
            filter_dict["id"] = {"$nin": filters.exclude_relationship_ids}

        if filters.document_id:
            if isinstance(filters.document_id, list):
                if len(filters.document_id) != 0:
                    filter_dict["document_id"] = {"$in": filters.document_id}
            else:
                filter_dict["document_id"] = {"$eq": filters.document_id}

        if filters.chunk_id:
            if isinstance(filters.chunk_id, list):
                if len(filters.chunk_id) != 0:
                    filter_dict["chunk_id"] = {"$in": filters.chunk_id}
            else:
                filter_dict["chunk_id"] = {"$eq": filters.chunk_id}

        if filters.metadata:
            for key, value in filters.metadata.items():
                op = "$in" if isinstance(value, list) else "$eq"
                filter_dict[f"meta.{key}"] = {op: value}

        return filter_dict

    def create_relationship(
        self,
        source_entity: Entity | UUID,
        target_entity: Entity | UUID,
        description: Optional[str] = None,
        meta: Optional[dict] = {},
        embedding: Optional[Any] = None,
    ) -> Relationship:
        """
        Create a relationship between two entities.
        """
        if isinstance(source_entity, UUID):
            source_entity = self.get_entity(source_entity)
        if isinstance(target_entity, UUID):
            target_entity = self.get_entity(target_entity)

        if embedding is None:
            embedding = self._get_relationship_embedding(
                source_entity.name,
                source_entity.description,
                target_entity.name,
                target_entity.description,
                description,
            )

        relationship = self._relationship_db_model(
            source_entity_id=source_entity.id,
            target_entity_id=target_entity.id,
            description=description,
            meta=meta,
            embedding=embedding,
        )
        return self._relationship_table.insert(relationship)

    def _get_relationship_embedding(
        self,
        source_entity_name: str,
        source_entity_description,
        target_entity_name: str,
        target_entity_description: str,
        relationship_desc: str,
    ) -> List[float]:
        embedding_str = (
            f"{source_entity_name}({source_entity_description}) -> "
            f"{relationship_desc} -> {target_entity_name}({target_entity_description}) "
        )
        return self._embedding_model.get_text_embedding(embedding_str)

    def update_relationship(
        self, relationship: Relationship | UUID, update: RelationshipUpdate
    ) -> Relationship:
        if isinstance(relationship, UUID):
            relationship = self.get_relationship(relationship)

        update_dict = update.model_dump()
        if update.embedding is None:
            update_dict["embedding"] = self._get_relationship_embedding(
                relationship.source_entity.name,
                relationship.source_entity.description,
                relationship.target_entity.name,
                relationship.target_entity.description,
                relationship.description,
            )

        self._relationship_table.update(
            values=update_dict, filters={"id": relationship.id}
        )
        # FIXME: pytidb should return the updated relationship.
        relationship = self._relationship_table.get(relationship.id)
        return relationship

    def delete_relationship(self, relationship_id: UUID):
        return self._relationship_table.delete(filters={"id": relationship_id})

    # Knowledge Graph Operations

    def add(self, knowledge_graph: KnowledgeGraphCreate) -> Optional[KnowledgeGraph]:
        with self._db.session():
            # Create or find entities
            entity_map = {}
            for entity in knowledge_graph.entities:
                created_entity = self.find_or_create_entity(
                    entity_type=EntityType.original,
                    name=entity.name,
                    description=entity.description,
                    meta=entity.meta,
                )
                entity_map[entity.name] = created_entity
            entities = list(entity_map.values())

            # Create relationships
            relationships = []
            for rel in knowledge_graph.relationships:
                logger.info("Saving relationship: %s", rel.description)
                source_entity = entity_map.get(rel.source_entity_name)
                if not source_entity:
                    logger.warning(
                        "Source entity not found for relationship: %s", str(rel)
                    )
                    continue

                target_entity = entity_map.get(rel.target_entity_name)
                if not target_entity:
                    logger.warning(
                        "Target entity not found for relationship: %s", str(rel)
                    )
                    continue

                relationship = self.create_relationship(
                    source_entity=source_entity,
                    target_entity=target_entity,
                    description=rel.description,
                    meta=rel.meta,
                )
                relationships.append(relationship)

        return KnowledgeGraph(
            entities=[Entity(**entity.model_dump()) for entity in entities],
            relationships=[
                Relationship(**relationship.model_dump())
                for relationship in relationships
            ],
        )

    # Graph Store Operations

    def reset(self):
        with self._db.session():
            self._db.execute("SET FOREIGN_KEY_CHECKS = 0")
            self._relationship_table.truncate()
            self._entity_table.truncate()
            self._db.execute("SET FOREIGN_KEY_CHECKS = 1")

    def recreate(self):
        self._db.drop_table(self._relationship_table.table_name)
        self._db.drop_table(self._entity_table.table_name)
        self._entity_table = self._db.create_table(schema=self._entity_db_model)
        self._relationship_table = self._db.create_table(
            schema=self._relationship_db_model
        )

    def drop(self):
        self._db.drop_table(self._relationship_table.table_name)
        self._db.drop_table(self._entity_table.table_name)



from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel, Field
from autoflow.storage.graph_store.types import (
    Entity,
    EntityCreate,
    KnowledgeGraphCreate,
    Relationship,
    RelationshipCreate,
)


# Generated Knowledge Graph


class GeneratedEntity(BaseModel):
    name: str
    description: str
    meta: Dict[str, Any] = Field(default_factory=dict)


class GeneratedRelationship(BaseModel):
    source_entity_name: str
    target_entity_name: str
    description: str
    meta: Dict[str, Any] = Field(default_factory=dict)


class GeneratedKnowledgeGraph(BaseModel):
    entities: List[GeneratedEntity]
    relationships: List[GeneratedRelationship]

    def to_create(
        self,
        chunk_id: Optional[UUID] = None,
        document_id: Optional[UUID] = None,
    ) -> KnowledgeGraphCreate:
        return KnowledgeGraphCreate(
            entities=[
                EntityCreate(
                    name=e.name,
                    description=e.description,
                    meta=e.meta,
                )
                for e in self.entities
            ],
            relationships=[
                RelationshipCreate(
                    source_entity_name=r.source_entity_name,
                    target_entity_name=r.target_entity_name,
                    description=r.description,
                    meta=r.meta,
                    weight=0,
                    chunk_id=chunk_id,
                    document_id=document_id,
                )
                for r in self.relationships
            ],
        )


# Retrieved Knowledge Graph


class RetrievedEntity(Entity):
    similarity_score: Optional[float] = Field(default=None)
    score: Optional[float] = Field(default=None)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other: "RetrievedEntity"):
        return self.id == other.id


class RetrievedRelationship(Relationship):
    similarity_score: Optional[float] = Field(default=None)
    score: Optional[float] = Field(default=None)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other: "RetrievedRelationship"):
        return self.id == other.id


class RetrievedKnowledgeGraph(BaseModel):
    query: Optional[str] = Field(
        description="The query used to retrieve the knowledge graph",
        default=None,
    )
    entities: List[Entity] = Field(
        description="List of entities in the knowledge graph", default_factory=list
    )
    relationships: List[RetrievedRelationship] = Field(
        description="List of relationships in the knowledge graph", default_factory=list
    )



from collections import defaultdict
from typing import List, Set, Tuple, Optional

from autoflow.knowledge_graph.types import (
    RetrievedKnowledgeGraph,
    RetrievedRelationship,
)
from autoflow.models.embedding_models import EmbeddingModel
from autoflow.storage.graph_store import GraphStore
from autoflow.knowledge_graph.retrievers.base import KGRetriever
from autoflow.storage.graph_store.types import (
    Entity,
    EntityDegree,
    EntityFilters,
    Relationship,
    EntityType,
    RelationshipFilters,
)
from autoflow.storage.types import QueryBundle


# The configuration for the weight coefficient
# format: ((min_weight, max_weight), coefficient)
DEFAULT_WEIGHT_COEFFICIENTS = [
    ((0, 100), 0.01),
    ((100, 1000), 0.001),
    ((1000, 10000), 0.0001),
    ((10000, float("inf")), 0.00001),
]

# The configuration for the range search
# format: ((min_distance, max_distance), search_ratio)
# The sum of search ratio should be 1 except some case we want to search as many as possible relationships.
# In this case, we set the search ratio to 1, and the other search ratio sum should be 1
DEFAULT_RANGE_SEARCH_CONFIG = [
    ((0.0, 0.25), 1),
    ((0.25, 0.35), 0.7),
    ((0.35, 0.45), 0.2),
    ((0.45, 0.55), 0.1),
]

DEFAULT_DEGREE_COEFFICIENT = 0.001


class WeightedGraphRetriever(KGRetriever):
    def __init__(
        self,
        kg_store: GraphStore,
        embedding_model: EmbeddingModel,
        with_degree: bool = False,
        alpha: float = 1,
        weight_coefficients: List[Tuple[float, float]] = None,
        search_range_config: List[Tuple[Tuple[float, float], float]] = None,
        degree_coefficient: float = DEFAULT_DEGREE_COEFFICIENT,
        fetch_synopsis_entities_num: int = 2,
        max_neighbors: int = 10,
    ):
        super().__init__(kg_store)
        self._embedding_model = embedding_model
        self.with_degree = with_degree
        self.alpha = alpha
        self.weight_coefficients = weight_coefficients or DEFAULT_WEIGHT_COEFFICIENTS
        self.search_range_config = search_range_config or DEFAULT_RANGE_SEARCH_CONFIG
        self.degree_coefficient = degree_coefficient
        self.fetch_synopsis_entities_num = fetch_synopsis_entities_num
        self.max_neighbors = max_neighbors

    def retrieve(
        self,
        query: str,
        depth: int = 2,
        metadata_filters: Optional[dict] = None,
    ) -> RetrievedKnowledgeGraph:
        query_embedding = self._embedding_model.get_query_embedding(query)

        visited_relationships = set()
        visited_entities = set()

        new_relationships = self._weighted_search_relationships(
            query_embedding=query_embedding,
            visited_relationships=visited_relationships,
            visited_entities=visited_entities,
            metadata_filters=metadata_filters,
        )

        if len(new_relationships) == 0:
            return RetrievedKnowledgeGraph(
                entities=[],
                relationships=[],
            )

        for rel, score in new_relationships:
            visited_relationships.add(
                RetrievedRelationship(
                    **rel.model_dump(),
                    similarity_score=score,
                    score=score,
                )
            )
            visited_entities.add(rel.source_entity)
            visited_entities.add(rel.target_entity)

        for _ in range(depth - 1):
            actual_number = 0
            progress = 0
            for search_config in DEFAULT_RANGE_SEARCH_CONFIG:
                search_ratio = search_config[1]
                search_distance_range = search_config[0]
                remaining_number = self.max_neighbors - actual_number
                # calculate the expected number based search progress
                # It's an accumulative search, so the expected number should be the difference between the expected number and the actual number
                expected_number = (
                    int((search_ratio + progress) * self.max_neighbors - actual_number)
                    if progress * self.max_neighbors > actual_number
                    else int(search_ratio * self.max_neighbors)
                )
                if expected_number > remaining_number:
                    expected_number = remaining_number
                if remaining_number <= 0:
                    break

                new_relationships = self._weighted_search_relationships(
                    query_embedding=query_embedding,
                    visited_relationships=visited_relationships,
                    visited_entities=visited_entities,
                    search_distance_range=search_distance_range,
                    top_k=expected_number,
                    metadata_filters=metadata_filters,
                )

                for rel, score in new_relationships:
                    visited_relationships.add(
                        RetrievedRelationship(
                            **rel.model_dump(),
                            similarity_score=score,
                            score=score,
                        )
                    )
                    visited_entities.add(rel.source_entity)
                    visited_entities.add(rel.target_entity)

                actual_number += len(new_relationships)
                # search_ratio == 1 won't count the progress
                if search_ratio != 1:
                    progress += search_ratio

        # Fetch related synopsis entities.
        synopsis_entities = self._kg_store.search_entities(
            query=QueryBundle(query_embedding=query_embedding),
            top_k=self.fetch_synopsis_entities_num,
            filters=EntityFilters(
                entity_type=EntityType.synopsis,
            ),
        )
        if len(synopsis_entities) > 0:
            visited_entities.update(synopsis_entities)

        # Rerank final relationships.
        return_relationships = list(visited_relationships)
        return_relationships.sort(key=lambda x: x.score, reverse=True)
        self._fill_entity(return_relationships)

        return_entities = [
            Entity(**e.model_dump())
            for e in visited_entities
        ]

        return RetrievedKnowledgeGraph(
            entities=return_entities,
            relationships=return_relationships,
        )

    def _fill_entity(self, relationships: List[RetrievedRelationship]):
        # FIXME: pytidb should return the relationship field: target_entity, source_entity.
        entity_ids = [item.target_entity_id for item in relationships]
        entity_ids.extend([item.source_entity_id for item in relationships])
        entities = self._kg_store.list_entities(
            filters=EntityFilters(entity_id=entity_ids)
        )
        entity_map = {entity.id: entity for entity in entities}
        for rel in relationships:
            rel.target_entity = Entity(**entity_map[rel.target_entity_id].model_dump())
            rel.source_entity = Entity(**entity_map[rel.source_entity_id].model_dump())

    def _weighted_search_relationships(
        self,
        query_embedding: List[float],
        visited_relationships: Set[RetrievedRelationship],
        visited_entities: Set[Entity],
        search_distance_range: Tuple[float, float] = (0, 1),
        top_k: int = 10,
        metadata_filters: Optional[dict] = None,
    ) -> List[RetrievedRelationship]:
        visited_entity_ids = [e.id for e in visited_entities]
        visited_relationship_ids = [r.id for r in visited_relationships]
        relationships_with_score = self._kg_store.search_relationships(
            query=QueryBundle(query_embedding=query_embedding),
            filters=RelationshipFilters(
                source_entity_id=visited_entity_ids,
                exclude_relationship_ids=visited_relationship_ids,
                metadata=metadata_filters,
            ),
            distance_range=search_distance_range,
            top_k=top_k,
        )

        return self._rank_relationships(
            relationships_with_score=relationships_with_score,
            top_k=top_k,
        )

    def _rank_relationships(
        self,
        relationships_with_score: List[Tuple[Relationship, float]],
        top_k: int = 10,
    ) -> List[Tuple[Relationship, float]]:
        """
        Rerank the relationship based on distance and weight
        """
        # TODO: the degree can br pre-calc and stored in the database in advanced.
        if self.with_degree:
            entity_ids = set()
            for r, _ in relationships_with_score:
                entity_ids.add(r.source_entity_id)
                entity_ids.add(r.target_entity_id)
            entity_degrees = self._kg_store.bulk_calc_entities_degrees(entity_ids)
        else:
            entity_degrees = defaultdict(EntityDegree)

        reranked_relationships = []
        for r, similarity_score in relationships_with_score:
            embedding_distance = 1 - similarity_score
            source_in_degree = entity_degrees[r.source_entity_id].in_degree
            target_out_degree = entity_degrees[r.target_entity_id].out_degree
            final_score = self._calc_relationship_weighted_score(
                embedding_distance,
                r.weight,
                source_in_degree,
                target_out_degree,
            )
            reranked_relationships.append((r, final_score))

        # Rerank relationships based on the calculated score.
        reranked_relationships.sort(key=lambda x: x[1], reverse=True)
        return reranked_relationships[:top_k]

    def _calc_relationship_weighted_score(
        self,
        embedding_distance: float,
        weight: int = 0,
        in_degree: int = 0,
        out_degree: int = 0,
    ) -> float:
        weighted_score = self._calc_weight_score(weight)
        degree_score = 0
        if self.with_degree:
            degree_score = self._calc_degree_score(in_degree, out_degree)
        return self.alpha * (1 / embedding_distance) + weighted_score + degree_score

    def _calc_weight_score(self, weight: float) -> float:
        weight_score = 0.0
        remaining_weight = weight

        for weight_range, coefficient in self.weight_coefficients:
            if remaining_weight <= 0:
                break
            lower_bound, upper_bound = weight_range
            applicable_weight = min(upper_bound - lower_bound, remaining_weight)
            weight_score += applicable_weight * coefficient
            remaining_weight -= applicable_weight

        return weight_score

    def _calc_degree_score(self, in_degree: int, out_degree: int) -> float:
        return (in_degree - out_degree) * self.degree_coefficient



from typing import Optional, Dict

from pydantic import BaseModel, Field, model_validator

from autoflow.configs.models.providers import ModelProviders

DEFAULT_TEMPERATURE = 0.1


class LLMConfig(BaseModel):
    provider: ModelProviders = Field(
        description="Provider of the large language models (LLM) (e.g., 'openai')",
        default=ModelProviders.OPENAI,
    )
    config: Optional[Dict] = Field(
        description="Configuration for the specific database",
        default=None,
    )
    _llm_configs: Dict[str, str] = {
        "openai": "OpenAILLMConfig",
        "custom": "CustomLLMConfig",
    }

    @model_validator(mode="after")
    def validate_and_create_config(self) -> "LLMConfig":
        provider = self.provider.value
        config = self.config

        if provider not in self._llm_configs:
            raise ValueError(f"Unsupported llm provider: {provider}")

        module = __import__(
            f"autoflow.configs.models.llms.{provider}",
            fromlist=[self._llm_configs[provider]],
        )
        config_class = getattr(module, self._llm_configs[provider])

        if config is None:
            config = {}

        if not isinstance(config, dict):
            if not isinstance(config, config_class):
                raise ValueError(f"Invalid config type for llm provider: {provider}")
            return self

        self.config = config_class(**config)
        return self



import dspy

from autoflow.models.llms import LLM


def get_dspy_lm_by_llm(llm: LLM) -> dspy.LM:
    return dspy.LM(
        model=llm.model,
        max_tokens=llm.max_tokens,
        **llm.additional_kwargs,
    )



from typing import Optional, Literal, List

from pydantic import BaseModel, Field
from litellm import LlmProviders

ModelProviders = LlmProviders

ModelType = Literal["llm", "text_embedding", "rerank"]


class ModelProviderInfo(BaseModel):
    name: ModelProviders = Field(
        description="The name of the model provider.",
    )
    logo: Optional[str] = Field(
        description="The logo of the model provider", default=None
    )
    display_name: str = Field(
        description="The name of the model provider",
    )
    description: str = Field(
        description="The description of the model provider", default=None
    )
    website: Optional[str] = Field(
        description="The website of the model provider", default=None
    )
    supported_model_types: List[ModelType] = Field(
        description="The model types supported by the model provider"
    )


class ProviderConfig(BaseModel):
    api_key: Optional[str] = Field(
        title="API key",
        default=None,
    )
    api_base: Optional[str] = Field(
        title="API Base",
        default=None,
    )



import dspy

from autoflow.knowledge_graph.extractors.base import KGExtractor
from autoflow.knowledge_graph.programs.extract_covariates import (
    EntityCovariateExtractor,
)
from autoflow.knowledge_graph.programs.extract_graph import KnowledgeGraphExtractor
from autoflow.knowledge_graph.types import GeneratedKnowledgeGraph


class SimpleKGExtractor(KGExtractor):
    def __init__(self, dspy_lm: dspy.LM):
        super().__init__()
        self._dspy_lm = dspy_lm
        self._graph_extractor = KnowledgeGraphExtractor(dspy_lm)
        self._entity_metadata_extractor = EntityCovariateExtractor(dspy_lm)

    def extract(self, text: str) -> GeneratedKnowledgeGraph:
        knowledge_graph = self._graph_extractor.forward(text)
        knowledge_graph.entities = self._entity_metadata_extractor.forward(
            text, knowledge_graph.entities
        )
        return knowledge_graph



import uuid
from datetime import datetime
from typing import Optional

from pytidb.schema import TableModel, Field
from pytidb.datatype import DateTime
from pytidb.sql import func

from autoflow.utils import uuid6


class UUIDBaseModel(TableModel):
    id: uuid.UUID = Field(default_factory=uuid6.uuid7, primary_key=True)
    # Use sa_type instead of sa_column, refer to https://github.com/tiangolo/sqlmodel/discussions/743
    created_at: Optional[datetime] = Field(
        sa_type=DateTime(timezone=True),
        sa_column_kwargs={"server_default": func.now()},
    )
    updated_at: Optional[datetime] = Field(
        sa_type=DateTime(timezone=True),
        sa_column_kwargs={
            "server_default": func.now(),
            "onupdate": func.now(),
        },
    )



from llama_index.core.prompts.rich import RichPromptTemplate

QA_WITH_KNOWLEDGE_PROMPT_TEMPLATE = RichPromptTemplate(
    template_str="""
    {% chat role="system" %}
    We have provided context information below.
    ---------------------
    {% if knowledge_graph %}
    <knowledge_graph>
        <entities>
            {% for entity in knowledge_graph.entities %}
            <entity id="{{ entity.id }}">
            {{ entity.name }}: {{ entity.description }}
            </entity>
            {% endfor %}
        </entities>
        <relationships>
            {% for relationship in knowledge_graph.relationships %}
            <relationship id="{{ relationship.id }}">
            {{relationship.source_entity.name}} -> {{ relationship.description }} -> {{relationship.target_entity.name}}
            </relationship>
            {% endfor %}
        </relationships>
    </knowledge_graph>
    {% endif %}

    {% for chunk in chunks %}
    <chunk id="{{ chunk.id }}">
    {{ chunk.text }}
    </chunk>
    {% endfor %}
    ---------------------
    Given this information, please give a comprehensive answer to the question in Markdown format:
    {% endchat %}

    {% chat role="user" %}
    {{ query_str }}
    {% endchat %}
    """
)


from typing import Dict, Optional

from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.postprocessor.types import BaseNodePostprocessor

from autoflow.configs.models.providers import ModelProviders
from autoflow.configs.models.embeddings import EmbeddingModelConfig
from autoflow.configs.models.llms import LLMConfig
from autoflow.configs.models.providers.base import ProviderConfig
from autoflow.configs.models.rerankers import RerankerConfig

from autoflow.models.embedding_models import EmbeddingModel
from autoflow.models.llms import LLM
from autoflow.models.rerank_models import RerankModel


class ModelManager:
    _registry: Dict[ModelProviders, ProviderConfig] = {}

    @classmethod
    def load_from_db(cls):
        pass

    @classmethod
    def from_config(cls):
        pass

    def registry_provider(self, name: ModelProviders, config: ProviderConfig):
        self._registry[name] = config

    def get_provider_config(self, name: ModelProviders) -> Optional[ProviderConfig]:
        provider = self._registry.get(name)
        if provider is None:
            raise ValueError('Provider "{}" is not found.'.format(name))
        return provider

    def resolve_llm(
        self,
        provider: Optional[ModelProviders] = ModelProviders.OPENAI,
        config: Optional[Dict] = None,
    ) -> Optional[BaseLLM]:
        cfg = LLMConfig.model_validate(
            {
                "provider": provider,
                "config": config,
            }
        )
        provider_config = self.get_provider_config(cfg.provider)
        merged_config = {
            **provider_config.model_dump(),
            **cfg.config.model_dump(),
            "model": f"{cfg.provider.value}/{cfg.config.model}",
        }
        return LLM(**merged_config)

    def resolve_embedding_model(
        self,
        provider: Optional[ModelProviders] = ModelProviders.OPENAI,
        config: Optional[Dict] = None,
    ) -> Optional[BaseEmbedding]:
        cfg = EmbeddingModelConfig.model_validate(
            {
                "provider": provider,
                "config": config,
            }
        )

        provider_config = self.get_provider_config(cfg.provider)
        merged_config = {
            **provider_config.model_dump(),
            **cfg.config.model_dump(exclude={"model"}),
            "model_name": f"{cfg.provider.value}/{cfg.config.model}",
        }
        return EmbeddingModel(**merged_config)

    def resolve_rerank_model(
        self,
        provider: Optional[ModelProviders] = ModelProviders.OPENAI,
        config: Optional[Dict] = None,
    ) -> Optional[BaseNodePostprocessor]:
        cfg = RerankerConfig.model_validate(
            {
                "provider": provider,
                "config": config,
            }
        )
        provider_config = self.get_provider_config(cfg.provider)
        merged_config = {
            **provider_config.model_dump(),
            **cfg.config.model_dump(exclude={"model"}),
            "model": f"{cfg.provider.value}/{cfg.config.model}",
        }
        return RerankModel(**merged_config)


model_manager: ModelManager = ModelManager()



from typing import List, Optional
from sqlalchemy.engine import Engine

from autoflow.configs.db import DatabaseConfig
from autoflow.configs.knowledge_base import IndexMethod
from autoflow.configs.main import Config
from autoflow.db import get_db_engine_from_config
from autoflow.knowledge_base.base import KnowledgeBase
from autoflow.models.embedding_models import EmbeddingModel
from autoflow.models.llms import LLM
from autoflow.models.manager import ModelManager, model_manager as default_model_manager
from autoflow.models.rerank_models import RerankModel


class Autoflow:
    _db_engine = None

    def __init__(
        self,
        db_engine: Engine,
        model_manager: Optional[ModelManager] = None,
    ):
        self._db_engine = db_engine
        self._model_manager = model_manager or default_model_manager

    @classmethod
    def from_config(cls, config: Config) -> "Autoflow":
        db_engine = cls._init_db_engine(config.db)
        model_manager = ModelManager()
        return cls(db_engine=db_engine, model_manager=model_manager)

    @classmethod
    def _init_db_engine(cls, db_config: DatabaseConfig) -> Engine:
        if db_config.provider != "tidb":
            raise NotImplementedError(
                f"Unsupported database provider: {db_config.provider}."
            )
        return get_db_engine_from_config(db_config)

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    @property
    def llm_manager(self) -> "ModelManager":
        return self._model_manager

    def create_knowledge_base(
        self,
        name: str,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
        index_methods: Optional[List[IndexMethod]] = None,
        llm: Optional[LLM] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        rerank_model: Optional[RerankModel] = None,
    ):
        return KnowledgeBase(
            db_engine=self.db_engine,
            namespace=namespace,
            name=name,
            description=description,
            index_methods=index_methods,
            llm=llm,
            embedding_model=embedding_model,
            rerank_model=rerank_model,
        )



from typing import Optional

from pydantic import BaseModel, Field


class BaseEmbeddingConfig(BaseModel):
    model: str = Field(
        description="The model to use for the embedding",
        default="text-embedding-3-small",
    )
    max_tokens: Optional[int] = None
    temperature: float = 0.1
